"""
Test all things related to the ``jedi.api`` module.
"""

import os
from textwrap import dedent

import pytest
from pytest import raises
from parso import cache

from jedi import preload_module
from jedi.inference.gradual import typeshed
from test.helpers import test_dir, get_example_dir


def test_preload_modules():
    def check_loaded(*module_names):
        for grammar_cache in cache.parser_cache.values():
            if None in grammar_cache:
                break
        # Filter the typeshed parser cache.
        typeshed_cache_count = sum(
            1 for path in grammar_cache
            if path is not None and str(path).startswith(str(typeshed.TYPESHED_PATH))
        )
        # +1 for None module (currently used)
        assert len(grammar_cache) - typeshed_cache_count == len(module_names) + 1
        for i in module_names:
            assert [i in str(k) for k in grammar_cache.keys() if k is not None]

    old_cache = cache.parser_cache.copy()
    cache.parser_cache.clear()

    try:
        preload_module('sys')
        check_loaded()  # compiled (c_builtin) modules shouldn't be in the cache.
        preload_module('types', 'token')
        check_loaded('types', 'token')
    finally:
        cache.parser_cache.update(old_cache)


def test_empty_script(Script):
    assert Script('')


def test_line_number_errors(Script):
    """
    Script should raise a ValueError if line/column numbers are not in a
    valid range.
    """
    s = 'hello'
    # lines
    with raises(ValueError):
        Script(s).complete(2, 0)
    with raises(ValueError):
        Script(s).complete(0, 0)

    # columns
    with raises(ValueError):
        Script(s).infer(1, len(s) + 1)
    with raises(ValueError):
        Script(s).goto(1, -1)

    # ok
    Script(s).get_signatures(1, 0)
    Script(s).get_references(1, len(s))


def _check_number(Script, source, result='float'):
    completions = Script(source).complete()
    assert completions[0].parent().name == result


def test_completion_on_number_literals(Script):
    # No completions on an int literal (is a float).
    assert [c.name for c in Script('1. ').complete()] \
        == ['and', 'if', 'in', 'is', 'not', 'or']

    # Multiple points after an int literal basically mean that there's a float
    # and a call after that.
    _check_number(Script, '1..')
    _check_number(Script, '1.0.')

    # power notation
    _check_number(Script, '1.e14.')
    _check_number(Script, '1.e-3.')
    _check_number(Script, '9e3.')
    assert Script('1.e3..').complete() == []
    assert Script('1.e-13..').complete() == []


def test_completion_on_hex_literals(Script):
    assert Script('0x1..').complete() == []
    _check_number(Script, '0x1.', 'int')  # hexdecimal
    # Completing binary literals doesn't work if they are not actually binary
    # (invalid statements).
    assert Script('0b2.b').complete() == []
    _check_number(Script, '0b1.', 'int')  # binary

    _check_number(Script, '0x2e.', 'int')
    _check_number(Script, '0xE7.', 'int')
    _check_number(Script, '0xEa.', 'int')
    # theoretically, but people can just check for syntax errors:
    assert Script('0x.').complete() == []


def test_completion_on_complex_literals(Script):
    assert Script('1j..').complete() == []
    _check_number(Script, '1j.', 'complex')
    _check_number(Script, '44.j.', 'complex')
    _check_number(Script, '4.0j.', 'complex')
    # No dot no completion - I thought, but 4j is actually a literal after
    # which a keyword like or is allowed. Good times, haha!
    # However this has been disabled again, because it apparently annoyed
    # users. So no completion after j without a space :)
    assert not Script('4j').complete()
    assert ({c.name for c in Script('4j ').complete()}
            == {'if', 'and', 'in', 'is', 'not', 'or'})


def test_goto_non_name(Script, environment):
    assert Script('for').goto() == []

    assert Script('assert').goto() == []
    assert Script('True').goto() == []


def test_infer_on_non_name(Script):
    assert Script('import x').infer(column=0) == []


def test_infer_on_generator(Script, environment):
    script = Script('def x(): yield 1\ny=x()\ny')
    def_, = script.infer()
    assert def_.name == 'Generator'
    def_, = script.infer(only_stubs=True)
    assert def_.name == 'Generator'


def test_goto_definition_not_multiple(Script):
    """
    There should be only one result if it leads back to the same
    origin (e.g. instance method)
    """

    s = dedent('''\
            import random
            class A():
                def __init__(self, a):
                    self.a = 3

                def foo(self):
                    pass

            if random.randint(0, 1):
                a = A(2)
            else:
                a = A(1)
            a''')
    assert len(Script(s).infer()) == 1


def test_reference_description(Script):
    descs = [u.description for u in Script("foo = ''; foo").get_references()]
    assert set(descs) == {"foo = ''", 'foo'}


def test_get_line_code(Script):
    def get_line_code(source, line=None, **kwargs):
        # On Windows replace \r
        return Script(source).complete(line=line)[0].get_line_code(**kwargs).replace('\r', '')

    # On builtin
    assert get_line_code('abs') == 'def abs(__x: SupportsAbs[_T]) -> _T: ...\n'

    # On custom code
    first_line = 'def foo():\n'
    line = '    foo'
    code = first_line + line
    assert get_line_code(code) == first_line

    # With before/after
    code = code + '\nother_line'
    assert get_line_code(code, line=2) == first_line
    assert get_line_code(code, line=2, after=1) == first_line + line + '\n'
    assert get_line_code(code, line=2, after=2, before=1) == code
    # Should just be the whole thing, since there are no more lines on both
    # sides.
    assert get_line_code(code, line=2, after=3, before=3) == code


def test_get_line_code_on_builtin(Script, disable_typeshed):
    abs_ = Script('abs').complete()[0]
    assert abs_.name == 'abs'
    assert abs_.get_line_code() == ''
    assert abs_.line is None


def test_goto_follow_imports(Script):
    code = dedent("""
    import inspect
    inspect.isfunction""")
    definition, = Script(code).goto(column=0, follow_imports=True)
    assert definition.module_path.name == 'inspect.py'
    assert (definition.line, definition.column) == (1, 0)

    definition, = Script(code).goto(follow_imports=True)
    assert definition.module_path.name == 'inspect.py'
    assert (definition.line, definition.column) > (1, 0)

    code = '''def param(p): pass\nparam(1)'''
    start_pos = 1, len('def param(')

    script = Script(code)
    definition, = script.goto(*start_pos, follow_imports=True)
    assert (definition.line, definition.column) == start_pos
    assert definition.name == 'p'
    result, = definition.goto()
    assert result.name == 'p'
    result, = definition.infer()
    assert result.name == 'int'
    result, = result.infer()
    assert result.name == 'int'

    definition, = script.goto(*start_pos)
    assert (definition.line, definition.column) == start_pos

    d, = Script('a = 1\na').goto(follow_imports=True)
    assert d.name == 'a'


def test_goto_module(Script):
    def check(line, expected, follow_imports=False):
        script = Script(path=path)
        module, = script.goto(line=line, follow_imports=follow_imports)
        assert module.module_path == expected

    base_path = get_example_dir('simple_import')
    path = base_path.joinpath('__init__.py')

    check(1, base_path.joinpath('module.py'))
    check(1, base_path.joinpath('module.py'), follow_imports=True)
    check(5, base_path.joinpath('module2.py'))


def test_goto_definition_cursor(Script):

    s = ("class A():\n"
         "    def _something(self):\n"
         "        return\n"
         "    def different_line(self,\n"
         "                   b):\n"
         "        return\n"
         "A._something\n"
         "A.different_line"
         )

    in_name = 2, 9
    under_score = 2, 8
    cls = 2, 7
    should1 = 7, 10
    diff_line = 4, 10
    should2 = 8, 10

    def get_def(pos):
        return [d.description for d in Script(s).infer(*pos)]

    in_name = get_def(in_name)
    under_score = get_def(under_score)
    should1 = get_def(should1)
    should2 = get_def(should2)

    diff_line = get_def(diff_line)

    assert should1 == in_name
    assert should1 == under_score

    assert should2 == diff_line

    assert get_def(cls) == []


def test_no_statement_parent(Script):
    source = dedent("""
    def f():
        pass

    class C:
        pass

    variable = f if random.choice([0, 1]) else C""")
    defs = Script(source).infer(column=3)
    defs = sorted(defs, key=lambda d: d.line)
    assert [d.description for d in defs] == ['def f', 'class C']


def test_backslash_continuation_and_bracket(Script):
    code = dedent(r"""
    x = 0
    a = \
      [1, 2, 3, (x)]""")

    lines = code.splitlines()
    column = lines[-1].index('(')
    def_, = Script(code).infer(line=len(lines), column=column)
    assert def_.name == 'int'


def test_goto_follow_builtin_imports(Script):
    s = Script('import sys; sys')
    d, = s.goto(follow_imports=True)
    assert d.in_builtin_module() is True
    d, = s.goto(follow_imports=True, follow_builtin_imports=True)
    assert d.in_builtin_module() is True


def test_docstrings_for_completions(Script):
    for c in Script('').complete():
        assert isinstance(c.docstring(), str)


def test_fuzzy_completion(Script):
    script = Script('string =  "hello"\nstring.upper')
    assert ['isupper',
            'upper'] == [comp.name for comp in script.complete(fuzzy=True)]


def test_math_fuzzy_completion(Script, environment):
    script = Script('import math\nmath.og')
    expected = ['copysign', 'log', 'log10', 'log1p', 'log2']
    completions = script.complete(fuzzy=True)
    assert expected == [comp.name for comp in completions]
    for c in completions:
        assert c.complete is None


def test_file_fuzzy_completion(Script):
    path = os.path.join(test_dir, 'completion')
    script = Script('"{}/ep08_i'.format(path))
    expected = [
        'pep0484_basic.py"',
        'pep0484_generic_mismatches.py"',
        'pep0484_generic_parameters.py"',
        'pep0484_generic_passthroughs.py"',
        'pep0484_typing.py"',
    ]
    assert expected == [comp.name for comp in script.complete(fuzzy=True)]


@pytest.mark.parametrize(
    'code, column', [
        ('"foo"', 0),
        ('"foo"', 3),
        ('"foo"', None),
        ('"""foo"""', 5),
        ('"""foo"""', 1),
        ('"""foo"""', 2),
    ]
)
def test_goto_on_string(Script, code, column):
    script = Script(code)
    assert not script.infer(column=column)
    assert not script.goto(column=column)


def test_multi_goto(Script):
    script = Script('x = 1\ny = 1.0\nx\ny')
    x, = script.goto(line=3)
    y, = script.goto(line=4)
    assert x.line == 1
    assert y.line == 2


@pytest.mark.parametrize(
    'code, column, expected', [
        ('str() ', 3, 'str'),
        ('str() ', 4, 'str'),
        ('str() ', 5, 'str'),
        ('str() ', 6, None),
        ('str(    ) ', 6, None),
        ('   1', 1, None),
        ('str(1) ', 3, 'str'),
        ('str(1) ', 4, 'int'),
        ('str(1) ', 5, 'int'),
        ('str(1) ', 6, 'str'),
        ('str(1) ', 7, None),
        ('str( 1) ', 4, 'str'),
        ('str( 1) ', 5, 'int'),
        ('str(+1) ', 4, 'str'),
        ('str(+1) ', 5, 'int'),
        ('str(1, 1.) ', 3, 'str'),
        ('str(1, 1.) ', 4, 'int'),
        ('str(1, 1.) ', 5, 'int'),
        ('str(1, 1.) ', 6, None),
        ('str(1, 1.) ', 7, 'float'),
    ]
)
def test_infer_after_parentheses(Script, code, column, expected):
    completions = Script(code).infer(column=column)
    if expected is None:
        assert completions == []
    else:
        assert [c.name for c in completions] == [expected]



"""
Tests for `api.names`.
"""

from textwrap import dedent

import pytest


def _assert_definition_names(definitions, names):
    assert [d.name for d in definitions] == names


def _check_names(get_names, source, names):
    definitions = get_names(dedent(source))
    _assert_definition_names(definitions, names)
    return definitions


def test_get_definitions_flat(get_names):
    _check_names(get_names, """
        import module
        class Class:
            pass
        def func():
            pass
        data = None
        """, ['module', 'Class', 'func', 'data'])


def test_dotted_assignment(get_names):
    _check_names(get_names, """
    x = Class()
    x.y.z = None
    """, ['x', 'z'])  # TODO is this behavior what we want?


def test_multiple_assignment(get_names):
    _check_names(get_names, "x = y = None", ['x', 'y'])


def test_multiple_imports(get_names):
    _check_names(get_names, """
    from module import a, b
    from another_module import *
    """, ['a', 'b'])


def test_nested_definitions(get_names):
    definitions = _check_names(get_names, """
    class Class:
        def f():
            pass
        def g():
            pass
    """, ['Class'])
    subdefinitions = definitions[0].defined_names()
    _assert_definition_names(subdefinitions, ['f', 'g'])
    assert [d.full_name for d in subdefinitions] == ['__main__.Class.f', '__main__.Class.g']


def test_nested_class(get_names):
    definitions = _check_names(get_names, """
    class L1:
        class L2:
            class L3:
                def f(): pass
            def f(): pass
        def f(): pass
    def f(): pass
    """, ['L1', 'f'])
    subdefs = definitions[0].defined_names()
    subsubdefs = subdefs[0].defined_names()
    _assert_definition_names(subdefs, ['L2', 'f'])
    _assert_definition_names(subsubdefs, ['L3', 'f'])
    _assert_definition_names(subsubdefs[0].defined_names(), ['f'])


def test_class_fields_with_all_scopes_false(get_names):
    definitions = _check_names(get_names, """
    from module import f
    g = f(f)
    class C:
        h = g

    def foo(x=a):
       bar = x
       return bar
    """, ['f', 'g', 'C', 'foo'])
    C_subdefs = definitions[-2].defined_names()
    foo_subdefs = definitions[-1].defined_names()
    _assert_definition_names(C_subdefs, ['h'])
    _assert_definition_names(foo_subdefs, ['x', 'bar'])


def test_async_stmt_with_all_scopes_false(get_names):
    definitions = _check_names(get_names, """
    from module import f
    import asyncio

    g = f(f)
    class C:
        h = g
        def __init__(self):
            pass

        async def __aenter__(self):
            pass

    def foo(x=a):
       bar = x
       return bar

    async def async_foo(duration):
        async def wait():
            await asyncio.sleep(100)
        for i in range(duration//100):
            await wait()
        return duration//100*100

    async with C() as cinst:
        d = cinst
    """, ['f', 'asyncio', 'g', 'C', 'foo', 'async_foo', 'cinst', 'd'])
    C_subdefs = definitions[3].defined_names()
    foo_subdefs = definitions[4].defined_names()
    async_foo_subdefs = definitions[5].defined_names()
    cinst_subdefs = definitions[6].defined_names()
    _assert_definition_names(C_subdefs, ['h', '__init__', '__aenter__'])
    _assert_definition_names(foo_subdefs, ['x', 'bar'])
    _assert_definition_names(async_foo_subdefs, ['duration', 'wait', 'i'])
    # We treat d as a name outside `async with` block
    _assert_definition_names(cinst_subdefs, [])


def test_follow_imports(get_names):
    # github issue #344
    imp = get_names('import datetime')[0]
    assert imp.name == 'datetime'
    datetime_names = [str(d.name) for d in imp.defined_names()]
    assert 'timedelta' in datetime_names


def test_names_twice(get_names):
    code = dedent('''
    def lol():
        pass
    ''')

    defs = get_names(code)
    assert defs[0].defined_names() == []


def test_simple_name(get_names):
    defs = get_names('foo', references=True)
    assert not defs[0]._name.infer()


def test_no_error(get_names):
    code = dedent("""
        def foo(a, b):
            if a == 10:
                if b is None:
                    print("foo")
                a = 20
        """)
    func_name, = get_names(code)
    a, b, a20 = func_name.defined_names()
    assert a.name == 'a'
    assert b.name == 'b'
    assert a20.name == 'a'
    assert a20.goto() == [a20]


@pytest.mark.parametrize(
    'code, index, is_side_effect', [
        ('x', 0, False),
        ('x.x', 0, False),
        ('x.x', 1, False),
        ('x.x = 3', 0, False),
        ('x.x = 3', 1, True),
        ('def x(x): x.x = 3', 1, False),
        ('def x(x): x.x = 3', 3, True),
        ('import sys; sys.path', 0, False),
        ('import sys; sys.path', 1, False),
        ('import sys; sys.path', 2, False),
        ('import sys; sys.path = []', 2, True),
    ]
)
def test_is_side_effect(get_names, code, index, is_side_effect):
    names = get_names(code, references=True, all_scopes=True)
    assert names[index].is_side_effect() == is_side_effect


def test_no_defined_names(get_names):
    definition, = get_names("x = (1, 2)")

    assert not definition.defined_names()



"""
Test completions from *.pyc files:

 - generate a dummy python module
 - compile the dummy module to generate a *.pyc
 - delete the pure python dummy module
 - try jedi on the generated *.pyc
"""
import os
import shutil
import sys

import pytest

import jedi
from jedi.api.environment import SameEnvironment, InterpreterEnvironment


SRC = """class Foo:
    pass

class Bar:
    pass
"""


@pytest.fixture
def pyc_project_path(tmpdir):
    path = tmpdir.strpath
    dummy_package_path = os.path.join(path, "dummy_package")
    os.mkdir(dummy_package_path)
    with open(os.path.join(dummy_package_path, "__init__.py"), 'w', newline=''):
        pass

    dummy_path = os.path.join(dummy_package_path, 'dummy.py')
    with open(dummy_path, 'w', newline='') as f:
        f.write(SRC)
    import compileall
    compileall.compile_file(dummy_path)
    os.remove(dummy_path)

    # To import pyc modules, we must move them out of the __pycache__
    # directory and rename them to remove ".cpython-%s%d"
    # see: http://stackoverflow.com/questions/11648440/python-does-not-detect-pyc-files
    pycache = os.path.join(dummy_package_path, "__pycache__")
    for f in os.listdir(pycache):
        dst = f.replace('.cpython-%s%s' % sys.version_info[:2], "")
        dst = os.path.join(dummy_package_path, dst)
        shutil.copy(os.path.join(pycache, f), dst)
    try:
        yield path
    finally:
        shutil.rmtree(path)


@pytest.mark.parametrize('load_unsafe_extensions', [False, True])
def test_pyc(pyc_project_path, environment, load_unsafe_extensions):
    """
    The list of completion must be greater than 2.
    """
    path = os.path.join(pyc_project_path, 'blub.py')
    if not isinstance(environment, InterpreterEnvironment):
        # We are using the same version for pyc completions here, because it
        # was compiled in that version. However with interpreter environments
        # we also have the same version and it's easier to debug.
        environment = SameEnvironment()
    environment = environment
    project = jedi.Project(pyc_project_path, load_unsafe_extensions=load_unsafe_extensions)
    s = jedi.Script(
        "from dummy_package import dummy; dummy.",
        path=path,
        environment=environment,
        project=project,
    )
    if load_unsafe_extensions:
        assert len(s.complete()) >= 2
    else:
        assert not s.complete()



import textwrap
from typing import IO, Any, Callable, Dict, Generic, List, Optional, Text, Tuple, Type, TypeVar, Union, overload

_TB = TypeVar("_TB", bound="_BaseEntry")
_TP = TypeVar("_TP", bound="POFile")
_TM = TypeVar("_TM", bound="MOFile")

default_encoding: str

# wrapwidth: int
# encoding: str
# check_for_duplicates: bool
@overload
def pofile(pofile: Text, *, klass: Type[_TP], **kwargs: Any) -> _TP: ...
@overload
def pofile(pofile: Text, **kwargs: Any) -> POFile: ...
@overload
def mofile(mofile: Text, *, klass: Type[_TM], **kwargs: Any) -> _TM: ...
@overload
def mofile(mofile: Text, **kwargs: Any) -> MOFile: ...
def detect_encoding(file: Union[bytes, Text], binary_mode: bool = ...) -> str: ...
def escape(st: Text) -> Text: ...
def unescape(st: Text) -> Text: ...

class _BaseFile(List[_TB]):
    fpath: Text
    wrapwidth: int
    encoding: Text
    check_for_duplicates: bool
    header: Text
    metadata: Dict[Text, Text]
    metadata_is_fuzzy: bool
    def __init__(self, *args: Any, **kwargs: Any) -> None: ...
    def __unicode__(self) -> Text: ...
    def __contains__(self, entry: _TB) -> bool: ...  # type: ignore # AttributeError otherwise
    def __eq__(self, other: object) -> bool: ...
    def append(self, entry: _TB) -> None: ...
    def insert(self, index: int, entry: _TB) -> None: ...
    def metadata_as_entry(self) -> POEntry: ...
    def save(self, fpath: Optional[Text] = ..., repr_method: str = ...) -> None: ...
    def find(self, st: Text, by: str = ..., include_obsolete_entries: bool = ..., msgctxt: bool = ...) -> Optional[_TB]: ...
    def ordered_metadata(self) -> List[Text]: ...
    def to_binary(self) -> bytes: ...

class POFile(_BaseFile):
    def __unicode__(self) -> Text: ...
    def save_as_mofile(self, fpath: Text) -> None: ...
    def percent_translated(self) -> int: ...
    def translated_entries(self) -> List[POEntry]: ...
    def untranslated_entries(self) -> List[POEntry]: ...
    def fuzzy_entries(self) -> List[POEntry]: ...
    def obsolete_entries(self) -> List[POEntry]: ...
    def merge(self, refpot: POFile) -> None: ...

class MOFile(_BaseFile):
    MAGIC: int
    MAGIC_SWAPPED: int
    magic_number: Optional[int]
    version: int
    def __init__(self, *args: Any, **kwargs: Any) -> None: ...
    def save_as_pofile(self, fpath: str) -> None: ...
    def save(self, fpath: Optional[Text] = ...) -> None: ...  # type: ignore # binary file does not allow argument repr_method
    def percent_translated(self) -> int: ...
    def translated_entries(self) -> List[MOEntry]: ...
    def untranslated_entries(self) -> List[MOEntry]: ...
    def fuzzy_entries(self) -> List[MOEntry]: ...
    def obsolete_entries(self) -> List[MOEntry]: ...

class _BaseEntry(object):
    msgid: Text
    msgstr: Text
    msgid_plural: List[Text]
    msgstr_plural: List[Text]
    msgctxt: Text
    obsolete: bool
    encoding: str
    def __init__(self, *args: Any, **kwargs: Any) -> None: ...
    def __unicode__(self, wrapwidth: int = ...) -> Text: ...
    def __eq__(self, other: object) -> bool: ...

class POEntry(_BaseEntry):
    comment: Text
    tcomment: Text
    occurrences: List[Tuple[str, int]]
    flags: List[Text]
    previous_msgctxt: Optional[Text]
    previous_msgid: Optional[Text]
    previous_msgid_plural: Optional[Text]
    linenum: Optional[int]
    def __init__(self, *args: Any, **kwargs: Any) -> None: ...
    def __unicode__(self, wrapwidth: int = ...) -> Text: ...
    def __cmp__(self, other: POEntry) -> int: ...
    def __gt__(self, other: POEntry) -> bool: ...
    def __lt__(self, other: POEntry) -> bool: ...
    def __ge__(self, other: POEntry) -> bool: ...
    def __le__(self, other: POEntry) -> bool: ...
    def __eq__(self, other: Any) -> bool: ...
    def __ne__(self, other: Any) -> bool: ...
    def translated(self) -> bool: ...
    def merge(self, other: POEntry) -> None: ...
    @property
    def fuzzy(self) -> bool: ...
    @property
    def msgid_with_context(self) -> Text: ...
    def __hash__(self) -> int: ...

class MOEntry(_BaseEntry):
    comment: Text
    tcomment: Text
    occurrences: List[Tuple[str, int]]
    flags: List[Text]
    previous_msgctxt: Optional[Text]
    previous_msgid: Optional[Text]
    previous_msgid_plural: Optional[Text]
    def __init__(self, *args: Any, **kwargs: Any) -> None: ...
    def __hash__(self) -> int: ...

class _POFileParser(Generic[_TP]):
    fhandle: IO[Text]
    instance: _TP
    transitions: Dict[Tuple[str, str], Tuple[Callable[[], bool], str]]
    current_line: int
    current_entry: POEntry
    current_state: str
    current_token: Optional[str]
    msgstr_index: int
    entry_obsolete: int
    def __init__(self, pofile: Text, *args: Any, **kwargs: Any) -> None: ...
    def parse(self) -> _TP: ...
    def add(self, symbol: str, states: List[str], next_state: str) -> None: ...
    def process(self, symbol: str) -> None: ...
    def handle_he(self) -> bool: ...
    def handle_tc(self) -> bool: ...
    def handle_gc(self) -> bool: ...
    def handle_oc(self) -> bool: ...
    def handle_fl(self) -> bool: ...
    def handle_pp(self) -> bool: ...
    def handle_pm(self) -> bool: ...
    def handle_pc(self) -> bool: ...
    def handle_ct(self) -> bool: ...
    def handle_mi(self) -> bool: ...
    def handle_mp(self) -> bool: ...
    def handle_ms(self) -> bool: ...
    def handle_mx(self) -> bool: ...
    def handle_mc(self) -> bool: ...

class _MOFileParser(Generic[_TM]):
    fhandle: IO[bytes]
    instance: _TM
    def __init__(self, mofile: Text, *args: Any, **kwargs: Any) -> None: ...
    def __del__(self) -> None: ...
    def parse(self) -> _TM: ...

class TextWrapper(textwrap.TextWrapper):
    drop_whitespace: bool
    def __init__(self, *args: Any, **kwargs: Any) -> None: ...



from typing import Any, List, Optional, Union

from django.template.base import Origin
from django.template.engine import Engine

from .base import Loader as BaseLoader

class Loader(BaseLoader):
    dirs: Optional[List[str]] = ...
    def __init__(self, engine: Engine, dirs: Optional[List[str]] = ...) -> None: ...
    def get_dirs(self) -> Union[List[bytes], List[str]]: ...
    def get_contents(self, origin: Origin) -> Any: ...



"""
All character set and unicode related tests.
"""
from jedi import Project


def test_unicode_script(Script):
    s = "import datetime; datetime.timedelta"
    completions = Script(s).complete()
    assert len(completions)
    assert type(completions[0].description) is str

    s = "author='öä'; author"
    completions = Script(s).complete()
    x = completions[0].description
    assert type(x) is str

    s = "#-*- coding: iso-8859-1 -*-\nauthor='öä'; author"
    s = s.encode('latin-1')
    completions = Script(s).complete()
    assert type(completions[0].description) is str


def test_unicode_attribute(Script):
    """ github jedi-vim issue #94 """
    s1 = ('#-*- coding: utf-8 -*-\nclass Person():\n'
          '    name = "e"\n\nPerson().name.')
    completions1 = Script(s1).complete()
    assert 'strip' in [c.name for c in completions1]
    s2 = ('#-*- coding: utf-8 -*-\nclass Person():\n'
          '    name = "é"\n\nPerson().name.')
    completions2 = Script(s2).complete()
    assert 'strip' in [c.name for c in completions2]


def test_multibyte_script(Script):
    """ `jedi.Script` must accept multi-byte string source. """
    code = "import datetime; datetime.d"
    comment = "# multi-byte comment あいうえおä"
    s = ('%s\n%s') % (code, comment)
    assert len(Script(s).complete(1, len(code)))


def test_goto_definition_at_zero(Script):
    """Infer at zero sometimes raises issues."""
    assert Script("a").infer(1, 1) == []
    s = Script("str").infer(1, 1)
    assert len(s) == 1
    assert list(s)[0].description == 'class str'
    assert Script("").infer(1, 0) == []


def test_complete_at_zero(Script):
    s = Script("str").complete(1, 3)
    assert len(s) == 1
    assert list(s)[0].name == 'str'

    s = Script("").complete(1, 0)
    assert len(s) > 0


def test_wrong_encoding(Script, tmpdir):
    x = tmpdir.join('x.py')
    # Use both latin-1 and utf-8 (a really broken file).
    x.write_binary('foobar = 1\nä'.encode('latin-1') + 'ä'.encode('utf-8'))

    project = Project(tmpdir.strpath)
    c, = Script('import x; x.foo', project=project).complete()
    assert c.name == 'foobar'



"""
Named Params:
>>> def a(abc): pass
...
>>> a(abc=3)  # <- this stuff (abc)
"""

def a(abc):
    pass

#? 5 ['abc=']
a(abc)


def a(*some_args, **some_kwargs):
    pass

#? 11 []
a(some_args)

#? 13 []
a(some_kwargs)

def multiple(foo, bar):
    pass

#? 17 ['bar=']
multiple(foo, bar)

#? ['bar=']
multiple(foo, bar

my_lambda = lambda lambda_param: lambda_param + 1
#? 22 ['lambda_param=']
my_lambda(lambda_param)

# __call__ / __init__
class Test(object):
    def __init__(self, hello_other):
        pass

    def __call__(self, hello):
        pass

    def test(self, blub):
        pass

#? 10 ['hello_other=']
Test(hello=)
#? 12 ['hello=']
Test()(hello=)
#? 11 []
Test()(self=)
#? 16 []
Test().test(self=)
#? 16 ['blub=']
Test().test(blub=)

# builtins

#? 12 []
any(iterable=)


def foo(xyz):
    pass

#? 7 ['xyz=']
@foo(xy)
def x(): pass

#? 7 ['xyz=']
foo(xyz)
# No completion should be possible if it's not a simple name
#? 17 []
x = " "; foo(x.xyz)
#? 17 []
x = " "; foo([xyz)
#? 20 []
x = " "; foo(z[f,xyz)
#? 18 []
x = " "; foo(z[xyz)
#? 20 []
x = " "; foo(xyz[xyz)
#? 20 []
x = " "; foo(xyz[(xyz)

#? 8 ['xyz=']
@foo(xyz)
def x(): pass

@str
#? 8 ['xyz=']
@foo(xyz)
def x(): pass

# -----------------
# Only keyword arguments are valid
# -----------------

def x(bam, *, bar, baz):
    pass
def y(bam, *bal, bar, baz, **bag):
    pass
def z(bam, bar=2, *, bas=1):
    pass

#? 7 ['bar=', 'baz=']
x(1, ba)

#? 14 ['baz=']
x(1, bar=2, ba)
#? 7 ['bar=', 'baz=']
x(1, ba, baz=3)
#? 14 ['baz=']
x(1, bar=2, baz=3)
#? 7 ['BaseException']
x(basee)
#? 22 ['bar=', 'baz=']
x(1, 2, 3, 4, 5, 6, bar=2)

#? 14 ['baz=']
y(1, bar=2, ba)
#? 7 ['bar=', 'BaseException', 'baz=']
y(1, ba, baz=3)
#? 14 ['baz=']
y(1, bar=2, baz=3)
#? 7 ['BaseException']
y(basee)
#? 22 ['bar=', 'BaseException', 'baz=']
y(1, 2, 3, 4, 5, 6, bar=2)

#? 11 ['bar=', 'bas=']
z(bam=1, bar=2, bas=3)
#? 8 ['BaseException', 'bas=']
z(1, bas=2)
#? 12 ['BaseException']
z(1, bas=bas)

#? 19 ['dict']
z(1, bas=bas, **dic)
#? 18 ['dict']
z(1, bas=bas, *dic)



# For assignment expressions / named expressions / walrus operators / whatever
# they are called.

# python >= 3.8
b = (a:=1, a)

#? int()
b[0]
#?
b[1]

# Should not fail
b = ('':=1,)

#? int()
b[0]

def test_assignments():
    match = ''
    #? str()
    match
    #? 8 int()
    if match := 1:
        #? int()
        match
    #? int()
    match

def test_assignments2():
    class Foo:
        match = ''
    #? str()
    Foo.match
    #? 13 int()
    if Foo.match := 1:
        #? str()
        Foo.match
    #? str()
    Foo.match

    #?
    y
    #? 16 str()
    if y := Foo.match:
        #? str()
        y
    #? str()
    y

    #? 8 str()
    if z := Foo.match:
        pass



from textwrap import dedent
import inspect
from unittest import TestCase

import pytest

from jedi import cache
from jedi.parser_utils import get_signature
from jedi import Interpreter


def assert_signature(Script, source, expected_name, expected_index=0, line=None, column=None):
    signatures = Script(source).get_signatures(line, column)

    assert len(signatures) <= 1

    if not signatures:
        assert expected_name is None, \
            'There are no signatures, but `%s` expected.' % expected_name
    else:
        assert signatures[0].name == expected_name
        assert signatures[0].index == expected_index
        return signatures[0]


def test_valid_call(Script):
    assert_signature(Script, 'bool()', 'bool', column=5)


class TestSignatures(TestCase):
    @pytest.fixture(autouse=True)
    def init(self, Script):
        self.Script = Script

    def _run_simple(self, source, name, index=0, column=None, line=1):
        assert_signature(self.Script, source, name, index, line, column)

    def test_simple(self):
        run = self._run_simple

        # simple
        s1 = "tuple(a, bool("
        run(s1, 'tuple', 0, 6)
        run(s1, 'tuple', None, 8)
        run(s1, 'tuple', None, 9)
        run(s1, 'bool', 0, 14)

        s2 = "abs(), "
        run(s2, 'abs', 0, 4)
        run(s2, None, column=5)
        run(s2, None)

        s3 = "abs()."
        run(s3, None, column=5)
        run(s3, None)

    def test_more_complicated(self):
        run = self._run_simple

        s4 = 'abs(bool(), , set,'
        run(s4, None, column=3)
        run(s4, 'abs', 0, 4)
        run(s4, 'bool', 0, 9)
        run(s4, 'abs', 0, 10)
        run(s4, 'abs', None, 11)

        s5 = "tuple(1,\nif 2:\n def a():"
        run(s5, 'tuple', 0, 6)
        run(s5, 'tuple', None, 8)

        s6 = "bool().__eq__("
        run(s6, '__eq__', 0)
        run(s6, 'bool', 0, 5)

        s7 = "str().upper().center("
        s8 = "bool(int[abs("
        run(s7, 'center', 0)
        run(s8, 'abs', 0)
        run(s8, 'bool', 0, 10)

        run("import time; abc = time; abc.sleep(", 'sleep', 0)

    def test_issue_57(self):
        # jedi #57
        s = "def func(alpha, beta): pass\n" \
            "func(alpha='101',"
        self._run_simple(s, 'func', 0, column=13, line=2)

    def test_for(self):
        # jedi-vim #11
        self._run_simple("for tuple(", 'tuple', 0)
        self._run_simple("for s in tuple(", 'tuple', 0)


def test_with(Script):
    # jedi-vim #9
    sigs = Script("with open(").get_signatures()
    assert sigs
    assert all(sig.name == 'open' for sig in sigs)


def test_get_signatures_empty_parentheses_pre_space(Script):
    s = dedent("""\
    def f(a, b):
        pass
    f( )""")
    assert_signature(Script, s, 'f', 0, line=3, column=3)


def test_multiple_signatures(Script):
    s = dedent("""\
    if x:
        def f(a, b):
            pass
    else:
        def f(a, b):
            pass
    f(""")
    assert len(Script(s).get_signatures()) == 2


def test_get_signatures_whitespace(Script):
    # note: trailing space after 'abs'
    s = 'abs( \ndef x():\n    pass\n'
    assert_signature(Script, s, 'abs', 0, line=1, column=5)


def test_decorator_in_class(Script):
    """
    There's still an implicit param, with a decorator.
    Github issue #319.
    """
    s = dedent("""\
    def static(func):
        def wrapped(obj, *args):
            return f(type(obj), *args)
        return wrapped

    class C(object):
        @static
        def test(cls):
            return 10

    C().test(""")

    signatures = Script(s).get_signatures()
    assert len(signatures) == 1
    x = [p.description for p in signatures[0].params]
    assert x == ['param *args']


def test_additional_brackets(Script):
    assert_signature(Script, 'abs((', 'abs', 0)


def test_unterminated_strings(Script):
    assert_signature(Script, 'abs(";', 'abs', 0)


def test_whitespace_before_bracket(Script):
    assert_signature(Script, 'abs (', 'abs', 0)
    assert_signature(Script, 'abs (";', 'abs', 0)
    assert_signature(Script, 'abs\n(', None)


def test_brackets_in_string_literals(Script):
    assert_signature(Script, 'abs (" (', 'abs', 0)
    assert_signature(Script, 'abs (" )', 'abs', 0)


def test_function_definitions_should_break(Script):
    """
    Function definitions (and other tokens that cannot exist within call
    signatures) should break and not be able to return a signature.
    """
    assert_signature(Script, 'abs(\ndef x', 'abs', 0)
    assert not Script('abs(\ndef x(): pass').get_signatures()


def test_flow_call(Script):
    assert not Script('if (1').get_signatures()


def test_chained_calls(Script):
    source = dedent('''
    class B():
      def test2(self, arg):
        pass

    class A():
      def test1(self):
        return B()

    A().test1().test2(''')

    assert_signature(Script, source, 'test2', 0)


def test_return(Script):
    source = dedent('''
    def foo():
        return '.'.join()''')

    assert_signature(Script, source, 'join', 0, column=len("    return '.'.join("))


def test_find_signature_on_module(Script):
    """github issue #240"""
    s = 'import datetime; datetime('
    # just don't throw an exception (if numpy doesn't exist, just ignore it)
    assert Script(s).get_signatures() == []


def test_complex(Script, environment):
    s = """
            def abc(a,b):
                pass

            def a(self):
                abc(

            if 1:
                pass
        """
    assert_signature(Script, s, 'abc', 0, line=6, column=20)
    s = """
            import re
            def huhu(it):
                re.compile(
                return it * 2
        """
    sig1, sig2 = sorted(Script(s).get_signatures(line=4, column=27), key=lambda s: s.line)
    assert sig1.name == sig2.name == 'compile'
    assert sig1.index == sig2.index == 0
    func1, = sig1._name.infer()
    func2, = sig2._name.infer()

    # Do these checks just for Python 3, I'm too lazy to deal with this
    # legacy stuff. ~ dave.
    assert get_signature(func1.tree_node) \
        == 'compile(pattern: AnyStr, flags: _FlagsType = ...) -> Pattern[AnyStr]'
    assert get_signature(func2.tree_node) \
        == 'compile(pattern: Pattern[AnyStr], flags: _FlagsType = ...) ->\nPattern[AnyStr]'

    # jedi-vim #70
    s = """def foo("""
    assert Script(s).get_signatures() == []

    # jedi-vim #116
    s = """import itertools; test = getattr(itertools, 'chain'); test("""
    assert_signature(Script, s, 'chain', 0)


def _params(Script, source, line=None, column=None):
    signatures = Script(source).get_signatures(line, column)
    assert len(signatures) == 1
    return signatures[0].params


def test_int_params(Script):
    sig1, sig2 = Script('int(').get_signatures()
    # int is defined as: `int(x[, base])`
    assert len(sig1.params) == 1
    assert sig1.params[0].name == 'x'
    assert len(sig2.params) == 2
    assert sig2.params[0].name == 'x'
    assert sig2.params[1].name == 'base'


def test_pow_params(Script):
    # See Github #1357.
    for sig in Script('pow(').get_signatures():
        param_names = [p.name for p in sig.params]
        assert param_names in (['base', 'exp'], ['base', 'exp', 'mod'])


def test_param_name(Script):
    sigs = Script('open(something,').get_signatures()
    for sig in sigs:
        # All of the signatures (in Python the function is overloaded),
        # contain the same param names.
        assert sig.params[0].name in ['file', 'name']
        assert sig.params[1].name == 'mode'
        assert sig.params[2].name == 'buffering'


def test_builtins(Script):
    """
    The self keyword should be visible even for builtins, if not
    instantiated.
    """
    p = _params(Script, 'str.endswith(')
    assert p[0].name == 'self'
    assert p[1].name == 'suffix'
    p = _params(Script, 'str().endswith(')
    assert p[0].name == 'suffix'


def test_signature_is_definition(Script):
    """
    Through inheritance, a signature is a sub class of Name.
    Check if the attributes match.
    """
    s = """class Spam(): pass\nSpam"""
    signature = Script(s + '(').get_signatures()[0]
    definition = Script(s + '(').infer(column=0)[0]
    signature.line == 1
    signature.column == 6

    # Now compare all the attributes that a Signature must also have.
    for attr_name in dir(definition):
        dont_scan = ['defined_names', 'parent', 'goto_assignments', 'infer',
                     'params', 'get_signatures', 'execute', 'goto',
                     'desc_with_module']
        if attr_name.startswith('_') or attr_name in dont_scan:
            continue

        attribute = getattr(definition, attr_name)
        signature_attribute = getattr(signature, attr_name)
        if inspect.ismethod(attribute):
            assert attribute() == signature_attribute()
        else:
            assert attribute == signature_attribute


def test_no_signature(Script):
    # str doesn't have a __call__ method
    assert Script('str()(').get_signatures() == []

    s = dedent("""\
    class X():
        pass
    X()(""")
    assert Script(s).get_signatures() == []
    assert len(Script(s).get_signatures(column=2)) == 1
    assert Script('').get_signatures() == []


def test_dict_literal_in_incomplete_call(Script):
    source = """\
    import json

    def foo():
        json.loads(

        json.load.return_value = {'foo': [],
                                  'bar': True}

        c = Foo()
    """

    script = Script(dedent(source))
    assert script.get_signatures(line=4, column=15)


def test_completion_interference(Script):
    """Seems to cause problems, see also #396."""
    cache.parser_cache.pop(None, None)
    assert Script('open(').get_signatures()

    # complete something usual, before doing the same get_signatures again.
    assert Script('from datetime import ').complete()

    assert Script('open(').get_signatures()


def test_keyword_argument_index(Script, environment):
    def get(source, column=None):
        return Script(source).get_signatures(column=column)[0]

    assert get('sorted([], key=a').index == 1
    assert get('sorted([], key=').index == 1
    assert get('sorted([], no_key=a').index is None

    kw_func = 'def foo(a, b): pass\nfoo(b=3, a=4)'
    assert get(kw_func, column=len('foo(b')).index == 0
    assert get(kw_func, column=len('foo(b=')).index == 1
    assert get(kw_func, column=len('foo(b=3, a=')).index == 0

    kw_func_simple = 'def foo(a, b): pass\nfoo(b=4)'
    assert get(kw_func_simple, column=len('foo(b')).index == 0
    assert get(kw_func_simple, column=len('foo(b=')).index == 1

    args_func = 'def foo(*kwargs): pass\n'
    assert get(args_func + 'foo(a').index == 0
    assert get(args_func + 'foo(a, b').index == 0

    kwargs_func = 'def foo(**kwargs): pass\n'
    assert get(kwargs_func + 'foo(a=2').index == 0
    assert get(kwargs_func + 'foo(a=2, b=2').index == 0

    both = 'def foo(*args, **kwargs): pass\n'
    assert get(both + 'foo(a=2').index == 1
    assert get(both + 'foo(a=2, b=2').index == 1
    assert get(both + 'foo(a=2, b=2)', column=len('foo(b=2, a=2')).index == 1
    assert get(both + 'foo(a, b, c').index == 0


code1 = 'def f(u, /, v=3, *, abc, abd, xyz): pass'
code2 = 'def g(u, /, v=3, *, abc, abd, xyz, **kwargs): pass'
code3 = 'def h(u, /, v, *args, x=1, y): pass'
code4 = 'def i(u, /, v, *args, x=1, y, **kwargs): pass'


_calls = [
    # No *args, **kwargs
    (code1, 'f(', 0),
    (code1, 'f(a', 0),
    (code1, 'f(a,', 1),
    (code1, 'f(a,b', 1),
    (code1, 'f(a,b,', 2),
    (code1, 'f(a,b,c', None),
    (code1, 'f(a,b,a', 2),
    (code1, 'f(a,b,a=', None),
    (code1, 'f(a,b,abc', 2),
    (code1, 'f(a,b,abc=(', 2),
    (code1, 'f(a,b,abc=(f,1,2,3', 2),
    (code1, 'f(a,b,abd', 3),
    (code1, 'f(a,b,x', 4),
    (code1, 'f(a,b,xy', 4),
    (code1, 'f(a,b,xyz=', 4),
    (code1, 'f(a,b,xy=', None),
    (code1, 'f(u=', (0, None)),
    (code1, 'f(v=', 1),

    # **kwargs
    (code2, 'g(a,b,a', 2),
    (code2, 'g(a,b,abc', 2),
    (code2, 'g(a,b,abd', 3),
    (code2, 'g(a,b,arr', 5),
    (code2, 'g(a,b,xy', 4),
    (code2, 'g(a,b,xyz=', 4),
    (code2, 'g(a,b,xy=', 5),
    (code2, 'g(a,b,abc=1,abd=4,', 4),
    (code2, 'g(a,b,abc=1,xyz=3,abd=4,', 5),
    (code2, 'g(a,b,abc=1,abd=4,lala', 5),
    (code2, 'g(a,b,abc=1,abd=4,lala=', 5),
    (code2, 'g(a,b,abc=1,abd=4,abd=', 5),
    (code2, 'g(a,b,kw', 5),
    (code2, 'g(a,b,kwargs=', 5),
    (code2, 'g(u=', (0, 5)),
    (code2, 'g(v=', 1),

    # *args
    (code3, 'h(a,b,c', 2),
    (code3, 'h(a,b,c,', 2),
    (code3, 'h(a,b,c,d', 2),
    (code3, 'h(a,b,c,d[', 2),
    (code3, 'h(a,b,c,(3,', 2),
    (code3, 'h(a,b,c,(3,)', 2),
    (code3, 'h(a,b,args=', None),
    (code3, 'h(u,v=', 1),
    (code3, 'h(u=', (0, None)),
    (code3, 'h(u,*xxx', 1),
    (code3, 'h(u,*xxx,*yyy', 1),
    (code3, 'h(u,*[]', 1),
    (code3, 'h(u,*', 1),
    (code3, 'h(u,*, *', 1),
    (code3, 'h(u,1,**', 3),
    (code3, 'h(u,**y', 1),
    (code3, 'h(u,x=2,**', 1),
    (code3, 'h(u,x=2,**y', 1),
    (code3, 'h(u,v=2,**y', 3),
    (code3, 'h(u,x=2,**vv', 1),

    # *args, **kwargs
    (code4, 'i(a,b,c,d', 2),
    (code4, 'i(a,b,c,d,e', 2),
    (code4, 'i(a,b,c,d,e=', 5),
    (code4, 'i(a,b,c,d,e=3', 5),
    (code4, 'i(a,b,c,d=,x=', 3),
    (code4, 'i(a,b,c,d=5,x=4', 3),
    (code4, 'i(a,b,c,d=5,x=4,y', 4),
    (code4, 'i(a,b,c,d=5,x=4,y=3,', 5),
    (code4, 'i(a,b,c,d=5,y=4,x=3,', 5),
    (code4, 'i(a,b,c,d=4,', 3),
    (code4, 'i(a,b,c,x=1,d=,', 4),

    # Error nodes
    (code4, 'i(1, [a,b', 1),
    (code4, 'i(1, [a,b=,', 2),
    (code4, 'i(1, [a?b,', 2),
    (code4, 'i(1, [a?b,*', 2),
    (code4, 'i(?b,*r,c', 1),
    (code4, 'i(?*', 0),
    (code4, 'i(?**', (0, 1)),

    # Random
    (code4, 'i(()', 0),
    (code4, 'i((),', 1),
    (code4, 'i([(),', 0),
    (code4, 'i([(,', 1),
    (code4, 'i(x,()', 1),
]


@pytest.mark.parametrize('ending', ['', ')'])
@pytest.mark.parametrize('code, call, expected_index', _calls)
def test_signature_index(Script, environment, code, call, expected_index, ending):
    if isinstance(expected_index, tuple):
        expected_index = expected_index[environment.version_info > (3, 8)]
    if environment.version_info < (3, 8):
        code = code.replace('/,', '')

    sig, = Script(code + '\n' + call + ending).get_signatures(column=len(call))
    index = sig.index
    assert expected_index == index


@pytest.mark.parametrize('code', ['foo', 'instance.foo'])
def test_arg_defaults(Script, environment, code):
    def foo(arg="bla", arg1=1):
        pass

    class Klass:
        def foo(self, arg="bla", arg1=1):
            pass

    instance = Klass()

    src = dedent("""
        def foo2(arg="bla", arg1=1):
            pass

        class Klass2:
            def foo2(self, arg="bla", arg1=1):
                pass

        instance = Klass2()
        """)

    executed_locals = dict()
    exec(src, None, executed_locals)
    locals_ = locals()

    def iter_scripts():
        yield Interpreter(code + '(', namespaces=[locals_])
        yield Script(src + code + "2(")
        yield Interpreter(code + '2(', namespaces=[executed_locals])

    for script in iter_scripts():
        signatures = script.get_signatures()
        assert signatures[0].params[0].description in ('param arg="bla"', "param arg='bla'")
        assert signatures[0].params[1].description == 'param arg1=1'


def test_bracket_start(Script):
    def bracket_start(src):
        signatures = Script(src).get_signatures()
        assert len(signatures) == 1
        return signatures[0].bracket_start

    assert bracket_start('abs(') == (1, 3)


def test_different_caller(Script):
    """
    It's possible to not use names, but another function result or an array
    index and then get the signature of it.
    """

    assert_signature(Script, '[abs][0](', 'abs', 0)
    assert_signature(Script, '[abs][0]()', 'abs', 0, column=len('[abs][0]('))

    assert_signature(Script, '(abs)(', 'abs', 0)
    assert_signature(Script, '(abs)()', 'abs', 0, column=len('(abs)('))


def test_in_function(Script):
    code = dedent('''\
    class X():
        @property
        def func(''')
    assert not Script(code).get_signatures()


def test_lambda_params(Script):
    code = dedent('''\
    my_lambda = lambda x: x+1
    my_lambda(1)''')
    sig, = Script(code).get_signatures(column=11)
    assert sig.index == 0
    assert sig.name == '<lambda>'
    assert [p.name for p in sig.params] == ['x']


CLASS_CODE = dedent('''\
class X():
    def __init__(self, foo, bar):
        self.foo = foo
''')


def test_class_creation(Script):

    sig, = Script(CLASS_CODE + 'X(').get_signatures()
    assert sig.index == 0
    assert sig.name == 'X'
    assert [p.name for p in sig.params] == ['foo', 'bar']


def test_call_init_on_class(Script):
    sig, = Script(CLASS_CODE + 'X.__init__(').get_signatures()
    assert [p.name for p in sig.params] == ['self', 'foo', 'bar']


def test_call_init_on_instance(Script):
    sig, = Script(CLASS_CODE + 'X().__init__(').get_signatures()
    assert [p.name for p in sig.params] == ['foo', 'bar']


def test_call_magic_method(Script):
    code = dedent('''\
    class X():
        def __call__(self, baz):
            pass
    ''')
    sig, = Script(code + 'X()(').get_signatures()
    assert sig.index == 0
    assert sig.name == 'X'
    assert [p.name for p in sig.params] == ['baz']

    sig, = Script(code + 'X.__call__(').get_signatures()
    assert [p.name for p in sig.params] == ['self', 'baz']
    sig, = Script(code + 'X().__call__(').get_signatures()
    assert [p.name for p in sig.params] == ['baz']


@pytest.mark.parametrize('column', [6, 9])
def test_cursor_after_signature(Script, column):
    source = dedent("""
        def foo(*args):
            pass
        foo()  # _
        """)

    script = Script(source)

    assert not script.get_signatures(4, column)


@pytest.mark.parametrize(
    'code, line, column, name, index', [
        ('abs(()\ndef foo(): pass', 1, None, 'abs', 0),
        ('abs(chr()  \ndef foo(): pass', 1, 10, 'abs', 0),
        ('abs(chr()\ndef foo(): pass', 1, None, 'abs', 0),
        ('abs(chr()\ndef foo(): pass', 1, 8, 'chr', 0),
        ('abs(chr()\ndef foo(): pass', 1, 7, 'abs', 0),
        ('abs(chr  ( \nclass y: pass', 1, None, 'chr', 0),
        ('abs(chr  ( \nclass y: pass', 1, 8, 'abs', 0),
        ('abs(chr  ( \nclass y: pass', 1, 9, 'abs', 0),
        ('abs(chr  ( \nclass y: pass', 1, 10, 'chr', 0),
    ]
)
def test_base_signatures(Script, code, line, column, name, index):
    sig, = Script(code).get_signatures(line=line, column=column)

    assert sig.name == name
    assert sig.index == index



import os
from textwrap import dedent
from pathlib import Path

import pytest

import jedi


@pytest.fixture()
def dir_with_content(tmpdir):
    with open(os.path.join(tmpdir.strpath, 'modx.py'), 'w', newline='') as f:
        f.write('import modx\nfoo\n')  # self reference
    return Path(tmpdir.strpath)


def test_rename_mod(Script, dir_with_content):
    script = Script(
        'import modx; modx\n',
        path=dir_with_content.joinpath('some_script.py'),
        project=jedi.Project(dir_with_content),
    )
    refactoring = script.rename(line=1, new_name='modr')
    refactoring.apply()

    p1 = dir_with_content.joinpath('modx.py')
    p2 = dir_with_content.joinpath('modr.py')
    expected_code = 'import modr\nfoo\n'
    assert not p1.exists()
    with open(p2, newline='') as f:
        assert f.read() == expected_code

    assert refactoring.get_renames() == [(p1, p2)]

    assert refactoring.get_changed_files()[p1].get_new_code() == expected_code

    assert refactoring.get_diff() == dedent('''\
        rename from modx.py
        rename to modr.py
        --- modx.py
        +++ modr.py
        @@ -1,3 +1,3 @@
        -import modx
        +import modr
         foo
        --- some_script.py
        +++ some_script.py
        @@ -1,2 +1,2 @@
        -import modx; modx
        +import modr; modr
        ''').format(dir=dir_with_content)


def test_rename_none_path(Script):
    refactoring = Script('foo', path=None).rename(new_name='bar')
    with pytest.raises(jedi.RefactoringError, match='on a Script with path=None'):
        refactoring.apply()
    assert refactoring


def test_diff_without_ending_newline(Script):
    refactoring = Script('a = 1\nb\na').rename(1, 0, new_name='c')
    assert refactoring.get_diff() == dedent('''\
        --- 
        +++ 
        @@ -1,3 +1,3 @@
        -a = 1
        +c = 1
         b
        -a
        +c
        ''')



# -----------------
# *args
# -----------------


def simple(a):
    return a


def nested(*args):
    return simple(*args)

nested(1)
#! 6 type-error-too-few-arguments
nested()


def nested_no_call_to_function(*args):
    return simple(1, *args)


def simple2(a, b, c):
    return b
def nested(*args):
    return simple2(1, *args)
def nested_twice(*args1):
    return nested(*args1)

nested_twice(2, 3)
#! 13 type-error-too-few-arguments
nested_twice(2)
#! 19 type-error-too-many-arguments
nested_twice(2, 3, 4)


# A named argument can be located before *args.
def star_args_with_named(*args):
    return simple2(c='', *args)

star_args_with_named(1, 2)
# -----------------
# **kwargs
# -----------------


def kwargs_test(**kwargs):
    return simple2(1, **kwargs)

kwargs_test(c=3, b=2)
#! 12 type-error-too-few-arguments
kwargs_test(c=3)
#! 12 type-error-too-few-arguments
kwargs_test(b=2)
#! 22 type-error-keyword-argument
kwargs_test(b=2, c=3, d=4)
#! 12 type-error-multiple-values
kwargs_test(b=2, c=3, a=4)


def kwargs_nested(**kwargs):
    return kwargs_test(b=2, **kwargs)

kwargs_nested(c=3)
#! 13 type-error-too-few-arguments
kwargs_nested()
#! 19 type-error-keyword-argument
kwargs_nested(c=2, d=4)
#! 14 type-error-multiple-values
kwargs_nested(c=2, a=4)
# TODO reenable
##! 14 type-error-multiple-values
#kwargs_nested(b=3, c=2)

# -----------------
# mixed *args/**kwargs
# -----------------

def simple_mixed(a, b, c):
    return b

def mixed(*args, **kwargs):
    return simple_mixed(1, *args, **kwargs)

mixed(1, 2)
mixed(1, c=2)
mixed(b=2, c=3)
mixed(c=4, b='')

# need separate functions, otherwise these might swallow the errors
def mixed2(*args, **kwargs):
    return simple_mixed(1, *args, **kwargs)


#! 7 type-error-too-few-arguments
mixed2(c=2)
#! 7 type-error-too-few-arguments
mixed2(3)
#! 13 type-error-too-many-arguments
mixed2(3, 4, 5)
# TODO reenable
##! 13 type-error-too-many-arguments
#mixed2(3, 4, c=5)
#! 7 type-error-multiple-values
mixed2(3, b=5)

# -----------------
# plain wrong arguments
# -----------------

#! 12 type-error-star-star
simple(1, **[])
#! 12 type-error-star-star
simple(1, **1)
class A(): pass
#! 12 type-error-star-star
simple(1, **A())

#! 11 type-error-star
simple(1, *1)



"""
These tests test Jedi's Parso usage. Basically there's not a lot of tests here,
because we're just checking if the API works. Bugfixes should be done in parso,
mostly.
"""

from textwrap import dedent

import pytest


@pytest.mark.parametrize(
    'code, line, column, until_line, until_column, message', [
        ('?\n', 1, 0, 1, 1, 'SyntaxError: invalid syntax'),
        ('x %% y', 1, 3, 1, 4, 'SyntaxError: invalid syntax'),
        ('"""\n\n', 1, 0, 3, 0, 'SyntaxError: EOF while scanning triple-quoted string literal'),
        ('(1, 2\n', 2, 0, 2, 0, 'SyntaxError: invalid syntax'),
        ('foo(1, 2\ndef x(): pass', 2, 0, 2, 3, 'SyntaxError: invalid syntax'),
    ]
)
def test_simple_syntax_errors(Script, code, line, column, until_line, until_column, message):
    e, = Script(code).get_syntax_errors()
    assert e.line == line
    assert e.column == column
    assert e.until_line == until_line
    assert e.until_column == until_column
    assert e.get_message() == message


@pytest.mark.parametrize(
    'code', [
        'x % y',
        'def x(x): pass',
        'def x(x):\n pass',
    ]
)
def test_no_syntax_errors(Script, code):
    assert not Script(code).get_syntax_errors()


def test_multi_syntax_error(Script):
    code = dedent('''\
        def x():
        1
        def y()
        1 + 1
        1 *** 3
        ''')
    x, y, power = Script(code).get_syntax_errors()
    assert x.line == 2
    assert x.column == 0
    assert y.line == 3
    assert y.column == 7
    assert power.line == 5
    assert power.column == 4



from typing import Optional

from cryptography.hazmat.backends.interfaces import CMACBackend
from cryptography.hazmat.primitives.ciphers import BlockCipherAlgorithm

class CMAC(object):
    def __init__(self, algorithm: BlockCipherAlgorithm, backend: Optional[CMACBackend] = ...) -> None: ...
    def copy(self) -> CMAC: ...
    def finalize(self) -> bytes: ...
    def update(self, data: bytes) -> None: ...
    def verify(self, signature: bytes) -> None: ...



from os.path import join
from itertools import chain
from functools import partial

import jedi
from ..helpers import test_dir


def test_import_empty(Script):
    """ github #340, return the full word. """
    completion = Script("import ").complete()[0]
    definition = completion.infer()[0]
    assert definition


def check_follow_definition_types(Script, source):
    # nested import
    completions = Script(source, path='some_path.py').complete()
    defs = chain.from_iterable(c.infer() for c in completions)
    return [d.type for d in defs]


def test_follow_import_incomplete(Script, environment):
    """
    Completion on incomplete imports should always take the full completion
    to do any type inference.
    """
    datetime = check_follow_definition_types(Script, "import itertool")
    assert datetime == ['module']

    # empty `from * import` parts
    itert = jedi.Script("from itertools import ").complete()
    definitions = [d for d in itert if d.name == 'chain']
    assert len(definitions) == 1
    assert [d.type for d in definitions[0].infer()] == ['class']

    # incomplete `from * import` part
    datetime = check_follow_definition_types(Script, "from datetime import datetim")
    assert set(datetime) == {'class', 'instance'}  # py3: builtin and pure py version
    # os.path check
    ospath = check_follow_definition_types(Script, "from os.path import abspat")
    assert set(ospath) == {'function'}

    # alias
    alias = check_follow_definition_types(Script, "import io as abcd; abcd")
    assert alias == ['module']


def test_follow_definition_nested_import(Script):
    Script = partial(Script, project=jedi.Project(join(test_dir, 'completion', 'import_tree')))
    types = check_follow_definition_types(Script, "import pkg.mod1; pkg")
    assert types == ['module']

    types = check_follow_definition_types(Script, "import pkg.mod1; pkg.mod1")
    assert types == ['module']

    types = check_follow_definition_types(Script, "import pkg.mod1; pkg.mod1.a")
    assert types == ['instance']


def test_follow_definition_land_on_import(Script):
    types = check_follow_definition_types(Script, "import datetime; datetim")
    assert types == ['module']



import gc
from pathlib import Path

from jedi import parser_utils
from parso import parse
from parso.cache import parser_cache
from parso.python import tree

import pytest


class TestCallAndName:
    def get_call(self, source):
        # Get the simple_stmt and then the first one.
        node = parse(source).children[0]
        if node.type == 'simple_stmt':
            return node.children[0]
        return node

    def test_name_and_call_positions(self):
        name = self.get_call('name\nsomething_else')
        assert name.value == 'name'
        assert name.start_pos == (1, 0)
        assert name.end_pos == (1, 4)

        leaf = self.get_call('1.0\n')
        assert leaf.value == '1.0'
        assert parser_utils.safe_literal_eval(leaf.value) == 1.0
        assert leaf.start_pos == (1, 0)
        assert leaf.end_pos == (1, 3)

    def test_call_type(self):
        call = self.get_call('hello')
        assert isinstance(call, tree.Name)

    def test_literal_type(self):
        literal = self.get_call('1.0')
        assert isinstance(literal, tree.Literal)
        assert type(parser_utils.safe_literal_eval(literal.value)) == float

        literal = self.get_call('1')
        assert isinstance(literal, tree.Literal)
        assert type(parser_utils.safe_literal_eval(literal.value)) == int

        literal = self.get_call('"hello"')
        assert isinstance(literal, tree.Literal)
        assert parser_utils.safe_literal_eval(literal.value) == 'hello'


def test_hex_values_in_docstring():
    source = r'''
        def foo(object):
            """
             \xff
            """
            return 1
        '''

    doc = parser_utils.clean_scope_docstring(next(parse(source).iter_funcdefs()))
    assert doc == '\xff'


@pytest.mark.parametrize(
    'code,signature', [
        ('def my_function(x, typed: Type, z):\n return', 'my_function(x, typed: Type, z)'),
        ('def my_function(x, y, z) -> str:\n return', 'my_function(x, y, z) -> str'),
        ('lambda x, y, z: x + y * z\n', '<lambda>(x, y, z)')
    ])
def test_get_signature(code, signature):
    node = parse(code, version='3.8').children[0]
    if node.type == 'simple_stmt':
        node = node.children[0]
    assert parser_utils.get_signature(node) == signature


def test_parser_cache_clear(Script):
    """
    If parso clears its cache, Jedi should not keep those resources, they
    should be freed.
    """
    script = Script("a = abs\na", path=Path(__file__).parent / 'parser_cache_test_foo.py')
    script.complete()
    module_id = id(script._module_node)
    del parser_cache[script._inference_state.grammar._hashed][script.path]
    del script

    gc.collect()
    assert module_id not in [id(m) for m in gc.get_referrers(tree.Module)]



from textwrap import dedent

import pytest
from parso import parse


def test_form_feed_characters(Script):
    s = "\f\nclass Test(object):\n    pass"
    Script(s).get_signatures(line=2, column=18)


def check_p(src):
    module_node = parse(src)
    assert src == module_node.get_code()
    return module_node


def test_if(Script):
    src = dedent('''\
    def func():
        x = 3
        if x:
            def y():
                return x
        return y()

    func()
    ''')

    # Two parsers needed, one for pass and one for the function.
    check_p(src)
    assert [d.name for d in Script(src).infer(8, 6)] == ['int']


def test_class_and_if(Script):
    src = dedent("""\
    class V:
        def __init__(self):
            pass

        if 1:
            c = 3

    def a_func():
        return 1

    # COMMENT
    a_func()""")
    check_p(src)
    assert [d.name for d in Script(src).infer()] == ['int']


def test_add_to_end(Script):
    """
    The diff parser doesn't parse everything again. It just updates with the
    help of caches, this is an example that didn't work.
    """

    a = dedent("""\
    class Abc():
        def abc(self):
            self.x = 3

    class Two(Abc):
        def g(self):
            self
    """)      # ^ here is the first completion

    b = "    def h(self):\n" \
        "        self."

    def complete(code, line=None, column=None):
        script = Script(code, path='example.py')
        assert script.complete(line, column)

    complete(a, 7, 12)
    complete(a + b)

    a = a[:-1] + '.\n'
    complete(a, 7, 13)
    complete(a + b)


def test_tokenizer_with_string_literal_backslash(Script):
    c = Script("statement = u'foo\\\n'; statement").infer()
    assert c[0]._name._value.get_safe_value() == 'foo'


def test_ellipsis_without_getitem(Script, environment):
    def_, = Script('x=...;x').infer()

    assert def_.name == 'ellipsis'



# python >= 3.6
from typing import List, Dict, overload, Tuple, TypeVar

lst: list
list_alias: List
list_str: List[str]
list_str: List[int]

# -------------------------
# With base classes
# -------------------------

@overload
def overload_f2(value: List) -> str: ...
@overload
def overload_f2(value: Dict) -> int: ...

#? str()
overload_f2([''])
#? int()
overload_f2({1.0: 1.0})
#? str()
overload_f2(lst)
#? str()
overload_f2(list_alias)
#? str()
overload_f2(list_str)


@overload
def overload_f3(value: list) -> str: ...
@overload
def overload_f3(value: dict) -> float: ...

#? str()
overload_f3([''])
#? float()
overload_f3({1.0: 1.0})
#? str()
overload_f3(lst)
#? str()
overload_f3(list_alias)
#? str()
overload_f3(list_str)

# -------------------------
# Generics Matching
# -------------------------

@overload
def overload_f1(value: List[str]) -> str: ...


@overload
def overload_f1(value: Dict[str, str]) -> Dict[str, str]: ...

def overload_f1():
    pass

#? str()
overload_f1([''])
#? str() dict()
overload_f1(1)
#? dict()
overload_f1({'': ''})

#? str() dict()
overload_f1(lst)
#? str() dict()
overload_f1(list_alias)
#? str()
overload_f1(list_str)
#? str() dict()
overload_f1(list_int)

# -------------------------
# Broken Matching
# -------------------------
T = TypeVar('T')

@overload
def broken_f1(value: 1) -> str: ...

@overload
def broken_f1(value: Tuple[T]) -> Tuple[T]: ...

tup: Tuple[float]
#? float()
broken_f1(broken_f1(tup))[0]



from typing import Any, Dict, Optional, Sequence, Set, Tuple, Union

from django.db.migrations.migration import Migration
from django.db.migrations.state import ProjectState

MIGRATIONS_MODULE_NAME: str

class MigrationLoader:
    connection: Any = ...
    disk_migrations: Dict[Tuple[str, str], Migration] = ...
    applied_migrations: Set[Tuple[str, str]] = ...
    ignore_no_migrations: bool = ...
    def __init__(self, connection: Any, load: bool = ..., ignore_no_migrations: bool = ...) -> None: ...
    @classmethod
    def migrations_module(cls, app_label: str) -> Tuple[Optional[str], bool]: ...
    unmigrated_apps: Set[str] = ...
    migrated_apps: Set[str] = ...
    def load_disk(self) -> None: ...
    def get_migration(self, app_label: str, name_prefix: str) -> Migration: ...
    def get_migration_by_prefix(self, app_label: str, name_prefix: str) -> Migration: ...
    def check_key(self, key: Tuple[str, str], current_app: str) -> Optional[Tuple[str, str]]: ...
    def add_internal_dependencies(self, key: Tuple[str, str], migration: Migration) -> None: ...
    def add_external_dependencies(self, key: Tuple[str, str], migration: Migration) -> None: ...
    graph: Any = ...
    replacements: Any = ...
    def build_graph(self) -> None: ...
    def check_consistent_history(self, connection: Any) -> None: ...
    def detect_conflicts(self) -> Dict[str, Set[str]]: ...
    def project_state(
        self, nodes: Optional[Union[Tuple[str, str], Sequence[Tuple[str, str]]]] = ..., at_end: bool = ...
    ) -> ProjectState: ...



# goto command tests are different in syntax

definition = 3
#! 0 ['a = definition']
a = definition

#! []
b
#! ['a = definition']
a

b = a
c = b
#! ['c = b']
c

cd = 1
#! 1 ['cd = c']
cd = c
#! 0 ['cd = e']
cd = e

#! ['module math']
import math
#! ['module math']
math

#! ['module math']
b = math
#! ['b = math']
b

#! 18 ['foo = 10']
foo = 10;print(foo)

# -----------------
# classes
# -----------------
class C(object):
    x = 3
    def b(self):
        #! ['b = math']
        b
        #! ['def b']
        self.b
        #! 14 ['def b']
        self.b()
        #! 14 ['def b']
        self.b.
        #! 11 ['param self']
        self.b
        #! ['x = 3']
        self.x
        #! 14 ['x = 3']
        self.x.
        return 1

    #! ['def b']
    b

#! ['b = math']
b

#! ['def b']
C.b
#! ['def b']
C().b
#! 0 ['class C']
C().b
#! 0 ['class C']
C().b

D = C
#! ['def b']
D.b
#! ['def b']
D().b

#! 0 ['D = C']
D().b
#! 0 ['D = C']
D().b

def c():
    return ''

#! ['def c']
c
#! 0 ['def c']
c()


class ClassVar():
    x = 3

#! ['x = 3']
ClassVar.x
#! ['x = 3']
ClassVar().x

# before assignments
#! 10 ['x = 3']
ClassVar.x = ''
#! 12 ['x = 3']
ClassVar().x = ''

# Recurring use of the same var name, github #315
def f(t=None):
    #! 9 ['param t=None']
    t = t or 1


class X():
    pass

#! 3 []
X(foo=x)


# Multiple inheritance
class Foo:
    def foo(self):
        print("foo")
class Bar:
    def bar(self):
        print("bar")
class Baz(Foo, Bar):
    def baz(self):
        #! ['def foo']
        super().foo
        #! ['def bar']
        super().bar
        #! ['instance Foo']
        super()

# -----------------
# imports
# -----------------

#! ['module import_tree']
import import_tree
#! ["a = ''"]
import_tree.a

#! ['module mod1']
import import_tree.mod1
#! ['module mod1']
from import_tree.mod1
#! ['a = 1']
import_tree.mod1.a

#! ['module pkg']
import import_tree.pkg
#! ['a = list']
import_tree.pkg.a

#! ['module mod1']
import import_tree.pkg.mod1
#! ['a = 1.0']
import_tree.pkg.mod1.a
#! ["a = ''"]
import_tree.a

#! ['module mod1']
from import_tree.pkg import mod1
#! ['a = 1.0']
mod1.a

#! ['module mod1']
from import_tree import mod1
#! ['a = 1']
mod1.a

#! ['a = 1.0']
from import_tree.pkg.mod1 import a

#! ['module os']
from .imports import os

#! ['some_variable = 1']
from . import some_variable

# -----------------
# anonymous classes
# -----------------
def func():
    class A():
        def b(self):
            return 1
    return A()

#! 8 ['def b']
func().b()

# -----------------
# on itself
# -----------------

#! 7 ['class ClassDef']
class ClassDef():
    """ abc """
    pass

# -----------------
# params
# -----------------

param = ClassDef
#! 8 ['param param']
def ab1(param): pass
#! 9 ['param param']
def ab2(param): pass
#! 11 ['param = ClassDef']
def ab3(a=param): pass

ab1(ClassDef);ab2(ClassDef);ab3(ClassDef)

# -----------------
# for loops
# -----------------

for i in range(1):
    #! ['for i in range(1): i']
    i

for key, value in [(1,2)]:
    #! ['for key, value in [(1,2)]: key']
    key

#! 4 ['for y in [1]: y']
for y in [1]:
    #! ['for y in [1]: y']
    y

# -----------------
# decorator
# -----------------
def dec(dec_param=3):
    pass

#! 8 ['param dec_param=3']
@dec(dec_param=5)
def y():
    pass

class ClassDec():
    def class_func(func):
        return func

#! 14 ['def class_func']
@ClassDec.class_func
def x():
    pass

#! 2 ['class ClassDec']
@ClassDec.class_func
def z():
    pass



import jedi
from jedi import debug

def test_simple():
    jedi.set_debug_function()
    debug.speed('foo')
    debug.dbg('bar')
    debug.warning('baz')
    jedi.set_debug_function(None, False, False, False)



"""
Test Jedi's operation understanding. Jedi should understand simple additions,
multiplications, etc.
"""
# -----------------
# numbers
# -----------------
x = [1, 'a', 1.0]

#? int() str() float()
x[12]

#? float()
x[1 + 1]

index = 0 + 1

#? str()
x[index]

#? int()
x[1 + (-1)]

def calculate(number):
    return number + constant

constant = 1

#? float()
x[calculate(1)]

def calculate(number):
    return number + constant

# -----------------
# strings
# -----------------

x = 'upp' + 'e'

#? str.upper
getattr(str, x + 'r')

a = "a"*3
#? str()
a
a = 3 * "a"
#? str()
a

a = 3 * "a"
#? str()
a

#? int()
(3 ** 3)
#? int()
(3 ** 'a')
#? int()
(3 + 'a')
#? bool()
(3 == 'a')
#? bool()
(3 >= 'a')

class X():
    foo = 2
#? int()
(X.foo ** 3)

# -----------------
# assignments
# -----------------

x = [1, 'a', 1.0]

i = 0
i += 1
i += 1
#? float()
x[i]

i = 1
i += 1
i -= 3
i += 1
#? int()
x[i]

# -----------------
# in
# -----------------

if 'X' in 'Y':
    a = 3
else:
    a = ''
# For now don't really check for truth values. So in should return both
# results.
#? str() int()
a

if 'X' not in 'Y':
    b = 3
else:
    b = ''
# For now don't really check for truth values. So in should return both
# results.
#? str() int()
b

# -----------------
# for flow assignments
# -----------------

class FooBar(object):
    fuu = 0.1
    raboof = 'fourtytwo'

# targets should be working
target = ''
for char in ['f', 'u', 'u']:
    target += char
#? float()
getattr(FooBar, target)

# github #24
target = u''
for char in reversed(['f', 'o', 'o', 'b', 'a', 'r']):
    target += char

#? str()
getattr(FooBar, target)


# -----------------
# repetition problems -> could be very slow and memory expensive - shouldn't
# be.
# -----------------

b = [str(1)]
l = list
for x in [l(0), l(1), l(2), l(3), l(4), l(5), l(6), l(7), l(8), l(9), l(10),
          l(11), l(12), l(13), l(14), l(15), l(16), l(17), l(18), l(19), l(20),
          l(21), l(22), l(23), l(24), l(25), l(26), l(27), l(28), l(29)]:
    b += x

#? str()
b[1]


# -----------------
# undefined names
# -----------------
a = foobarbaz + 'hello'

#? int() float()
{'hello': 1, 'bar': 1.0}[a]

# -----------------
# stubs
# -----------------

from datetime import datetime, timedelta

#?
(datetime - timedelta)
#? datetime()
(datetime() - timedelta())
#? timedelta()
(datetime() - datetime())
#? timedelta()
(timedelta() - datetime())
#? timedelta()
(timedelta() - timedelta())

# -----------------
# magic methods
# -----------------

class C:
    def __sub__(self, other) -> int: ...
    def __radd__(self, other) -> float: ...

#? int()
(C() - object())
#? C() object()
(object() - C())
#? C() object()
(C() + object())
#? float()
(object() + C())



import sys
from _typeshed import ReadableBuffer as ReadableBuffer
from typing import Any, Callable, Optional

from cryptography.hazmat.primitives.asymmetric.x25519 import X25519PrivateKey
from paramiko.message import Message
from paramiko.transport import Transport

if sys.version_info < (3, 0):
    from hashlib import _hash as _Hash
else:
    from hashlib import _Hash

c_MSG_KEXECDH_INIT: bytes
c_MSG_KEXECDH_REPLY: bytes

class KexCurve25519:
    hash_algo: Callable[[ReadableBuffer], _Hash]
    transport: Transport
    key: Optional[X25519PrivateKey]
    def __init__(self, transport: Transport) -> None: ...
    @classmethod
    def is_available(cls) -> bool: ...
    def start_kex(self) -> None: ...
    def parse_next(self, ptype: int, m: Message) -> None: ...



from email.mime.nonmultipart import MIMENonMultipart
from email.policy import Policy
from typing import Callable, Optional, Tuple, Union

_ParamsType = Union[str, None, Tuple[str, Optional[str], str]]

class MIMEAudio(MIMENonMultipart):
    def __init__(
        self,
        _audiodata: Union[str, bytes],
        _subtype: Optional[str] = ...,
        _encoder: Callable[[MIMEAudio], None] = ...,
        *,
        policy: Optional[Policy] = ...,
        **_params: _ParamsType,
    ) -> None: ...



from distutils.version import Version
from typing import Any, Callable, Dict, List, Optional, Protocol, Text, Tuple, Type, TypeVar, Union, overload

from click.core import Argument, Command, Context, Group, Option, Parameter, _ConvertibleType

_T = TypeVar("_T")
_F = TypeVar("_F", bound=Callable[..., Any])

class _IdentityFunction(Protocol):
    def __call__(self, __x: _T) -> _T: ...

_Callback = Callable[[Context, Union[Option, Parameter], Any], Any]

def pass_context(__f: _T) -> _T: ...
def pass_obj(__f: _T) -> _T: ...
def make_pass_decorator(object_type: type, ensure: bool = ...) -> _IdentityFunction: ...

# NOTE: Decorators below have **attrs converted to concrete constructor
# arguments from core.pyi to help with type checking.

def command(
    name: Optional[str] = ...,
    cls: Optional[Type[Command]] = ...,
    # Command
    context_settings: Optional[Dict[Any, Any]] = ...,
    help: Optional[str] = ...,
    epilog: Optional[str] = ...,
    short_help: Optional[str] = ...,
    options_metavar: str = ...,
    add_help_option: bool = ...,
    hidden: bool = ...,
    deprecated: bool = ...,
) -> Callable[[Callable[..., Any]], Command]: ...

# This inherits attrs from Group, MultiCommand and Command.

def group(
    name: Optional[str] = ...,
    cls: Type[Command] = ...,
    # Group
    commands: Optional[Dict[str, Command]] = ...,
    # MultiCommand
    invoke_without_command: bool = ...,
    no_args_is_help: Optional[bool] = ...,
    subcommand_metavar: Optional[str] = ...,
    chain: bool = ...,
    result_callback: Optional[Callable[..., Any]] = ...,
    # Command
    help: Optional[str] = ...,
    epilog: Optional[str] = ...,
    short_help: Optional[str] = ...,
    options_metavar: str = ...,
    add_help_option: bool = ...,
    hidden: bool = ...,
    deprecated: bool = ...,
    # User-defined
    **kwargs: Any,
) -> Callable[[Callable[..., Any]], Group]: ...
def argument(
    *param_decls: Text,
    cls: Type[Argument] = ...,
    # Argument
    required: Optional[bool] = ...,
    # Parameter
    type: Optional[_ConvertibleType] = ...,
    default: Optional[Any] = ...,
    callback: Optional[_Callback] = ...,
    nargs: Optional[int] = ...,
    metavar: Optional[str] = ...,
    expose_value: bool = ...,
    is_eager: bool = ...,
    envvar: Optional[Union[str, List[str]]] = ...,
    autocompletion: Optional[Callable[[Any, List[str], str], List[Union[str, Tuple[str, str]]]]] = ...,
) -> _IdentityFunction: ...
@overload
def option(
    *param_decls: Text,
    cls: Type[Option] = ...,
    # Option
    show_default: Union[bool, Text] = ...,
    prompt: Union[bool, Text] = ...,
    confirmation_prompt: bool = ...,
    hide_input: bool = ...,
    is_flag: Optional[bool] = ...,
    flag_value: Optional[Any] = ...,
    multiple: bool = ...,
    count: bool = ...,
    allow_from_autoenv: bool = ...,
    type: Optional[_ConvertibleType] = ...,
    help: Optional[Text] = ...,
    show_choices: bool = ...,
    # Parameter
    default: Optional[Any] = ...,
    required: bool = ...,
    callback: Optional[_Callback] = ...,
    nargs: Optional[int] = ...,
    metavar: Optional[str] = ...,
    expose_value: bool = ...,
    is_eager: bool = ...,
    envvar: Optional[Union[str, List[str]]] = ...,
    # User-defined
    **kwargs: Any,
) -> _IdentityFunction: ...
@overload
def option(
    *param_decls: str,
    cls: Type[Option] = ...,
    # Option
    show_default: Union[bool, Text] = ...,
    prompt: Union[bool, Text] = ...,
    confirmation_prompt: bool = ...,
    hide_input: bool = ...,
    is_flag: Optional[bool] = ...,
    flag_value: Optional[Any] = ...,
    multiple: bool = ...,
    count: bool = ...,
    allow_from_autoenv: bool = ...,
    type: _T = ...,
    help: Optional[str] = ...,
    show_choices: bool = ...,
    # Parameter
    default: Optional[Any] = ...,
    required: bool = ...,
    callback: Optional[Callable[[Context, Union[Option, Parameter], Union[bool, int, str]], _T]] = ...,
    nargs: Optional[int] = ...,
    metavar: Optional[str] = ...,
    expose_value: bool = ...,
    is_eager: bool = ...,
    envvar: Optional[Union[str, List[str]]] = ...,
    # User-defined
    **kwargs: Any,
) -> _IdentityFunction: ...
@overload
def option(
    *param_decls: str,
    cls: Type[Option] = ...,
    # Option
    show_default: Union[bool, Text] = ...,
    prompt: Union[bool, Text] = ...,
    confirmation_prompt: bool = ...,
    hide_input: bool = ...,
    is_flag: Optional[bool] = ...,
    flag_value: Optional[Any] = ...,
    multiple: bool = ...,
    count: bool = ...,
    allow_from_autoenv: bool = ...,
    type: Type[str] = ...,
    help: Optional[str] = ...,
    show_choices: bool = ...,
    # Parameter
    default: Optional[Any] = ...,
    required: bool = ...,
    callback: Callable[[Context, Union[Option, Parameter], str], Any] = ...,
    nargs: Optional[int] = ...,
    metavar: Optional[str] = ...,
    expose_value: bool = ...,
    is_eager: bool = ...,
    envvar: Optional[Union[str, List[str]]] = ...,
    # User-defined
    **kwargs: Any,
) -> _IdentityFunction: ...
@overload
def option(
    *param_decls: str,
    cls: Type[Option] = ...,
    # Option
    show_default: Union[bool, Text] = ...,
    prompt: Union[bool, Text] = ...,
    confirmation_prompt: bool = ...,
    hide_input: bool = ...,
    is_flag: Optional[bool] = ...,
    flag_value: Optional[Any] = ...,
    multiple: bool = ...,
    count: bool = ...,
    allow_from_autoenv: bool = ...,
    type: Type[int] = ...,
    help: Optional[str] = ...,
    show_choices: bool = ...,
    # Parameter
    default: Optional[Any] = ...,
    required: bool = ...,
    callback: Callable[[Context, Union[Option, Parameter], int], Any] = ...,
    nargs: Optional[int] = ...,
    metavar: Optional[str] = ...,
    expose_value: bool = ...,
    is_eager: bool = ...,
    envvar: Optional[Union[str, List[str]]] = ...,
    # User-defined
    **kwargs: Any,
) -> _IdentityFunction: ...
def confirmation_option(
    *param_decls: str,
    cls: Type[Option] = ...,
    # Option
    show_default: Union[bool, Text] = ...,
    prompt: Union[bool, Text] = ...,
    confirmation_prompt: bool = ...,
    hide_input: bool = ...,
    is_flag: bool = ...,
    flag_value: Optional[Any] = ...,
    multiple: bool = ...,
    count: bool = ...,
    allow_from_autoenv: bool = ...,
    type: Optional[_ConvertibleType] = ...,
    help: str = ...,
    show_choices: bool = ...,
    # Parameter
    default: Optional[Any] = ...,
    callback: Optional[_Callback] = ...,
    nargs: Optional[int] = ...,
    metavar: Optional[str] = ...,
    expose_value: bool = ...,
    is_eager: bool = ...,
    envvar: Optional[Union[str, List[str]]] = ...,
) -> _IdentityFunction: ...
def password_option(
    *param_decls: str,
    cls: Type[Option] = ...,
    # Option
    show_default: Union[bool, Text] = ...,
    prompt: Union[bool, Text] = ...,
    confirmation_prompt: bool = ...,
    hide_input: bool = ...,
    is_flag: Optional[bool] = ...,
    flag_value: Optional[Any] = ...,
    multiple: bool = ...,
    count: bool = ...,
    allow_from_autoenv: bool = ...,
    type: Optional[_ConvertibleType] = ...,
    help: Optional[str] = ...,
    show_choices: bool = ...,
    # Parameter
    default: Optional[Any] = ...,
    callback: Optional[_Callback] = ...,
    nargs: Optional[int] = ...,
    metavar: Optional[str] = ...,
    expose_value: bool = ...,
    is_eager: bool = ...,
    envvar: Optional[Union[str, List[str]]] = ...,
) -> _IdentityFunction: ...
def version_option(
    version: Optional[Union[str, Version]] = ...,
    *param_decls: str,
    cls: Type[Option] = ...,
    # Option
    prog_name: Optional[str] = ...,
    message: Optional[str] = ...,
    show_default: Union[bool, Text] = ...,
    prompt: Union[bool, Text] = ...,
    confirmation_prompt: bool = ...,
    hide_input: bool = ...,
    is_flag: bool = ...,
    flag_value: Optional[Any] = ...,
    multiple: bool = ...,
    count: bool = ...,
    allow_from_autoenv: bool = ...,
    type: Optional[_ConvertibleType] = ...,
    help: str = ...,
    show_choices: bool = ...,
    # Parameter
    default: Optional[Any] = ...,
    callback: Optional[_Callback] = ...,
    nargs: Optional[int] = ...,
    metavar: Optional[str] = ...,
    expose_value: bool = ...,
    is_eager: bool = ...,
    envvar: Optional[Union[str, List[str]]] = ...,
) -> _IdentityFunction: ...
def help_option(
    *param_decls: str,
    cls: Type[Option] = ...,
    # Option
    show_default: Union[bool, Text] = ...,
    prompt: Union[bool, Text] = ...,
    confirmation_prompt: bool = ...,
    hide_input: bool = ...,
    is_flag: bool = ...,
    flag_value: Optional[Any] = ...,
    multiple: bool = ...,
    count: bool = ...,
    allow_from_autoenv: bool = ...,
    type: Optional[_ConvertibleType] = ...,
    help: str = ...,
    show_choices: bool = ...,
    # Parameter
    default: Optional[Any] = ...,
    callback: Optional[_Callback] = ...,
    nargs: Optional[int] = ...,
    metavar: Optional[str] = ...,
    expose_value: bool = ...,
    is_eager: bool = ...,
    envvar: Optional[Union[str, List[str]]] = ...,
) -> _IdentityFunction: ...



from typing import Dict, Tuple, Callable

CacheValues = Tuple[str, str, str]
CacheValuesCallback = Callable[[], CacheValues]


_cache: Dict[str, Dict[str, CacheValues]] = {}


def save_entry(module_name: str, name: str, cache: CacheValues) -> None:
    try:
        module_cache = _cache[module_name]
    except KeyError:
        module_cache = _cache[module_name] = {}
    module_cache[name] = cache


def _create_get_from_cache(number: int) -> Callable[[str, str, CacheValuesCallback], str]:
    def _get_from_cache(module_name: str, name: str, get_cache_values: CacheValuesCallback) -> str:
        try:
            return _cache[module_name][name][number]
        except KeyError:
            v = get_cache_values()
            save_entry(module_name, name, v)
            return v[number]
    return _get_from_cache


get_type = _create_get_from_cache(0)
get_docstring_signature = _create_get_from_cache(1)
get_docstring = _create_get_from_cache(2)



import datetime
import decimal
import uuid

from django.db import models
from django.contrib.auth.models import User
from django.db.models.query_utils import DeferredAttribute


class TagManager(models.Manager):
    def specially_filtered_tags(self):
        return self.all()


class Tag(models.Model):
    tag_name = models.CharField()

    objects = TagManager()

    custom_objects = TagManager()


class Category(models.Model):
    category_name = models.CharField()


class AttachedData(models.Model):
    extra_data = models.TextField()


class BusinessModel(models.Model):
    attached_o2o = models.OneToOneField(AttachedData)

    category_fk = models.ForeignKey(Category)
    category_fk2 = models.ForeignKey('Category')
    category_fk3 = models.ForeignKey(1)
    category_fk4 = models.ForeignKey('models')
    category_fk5 = models.ForeignKey()

    integer_field = models.IntegerField()
    big_integer_field = models.BigIntegerField()
    positive_integer_field = models.PositiveIntegerField()
    small_integer_field = models.SmallIntegerField()
    char_field = models.CharField()
    text_field = models.TextField()
    email_field = models.EmailField()
    ip_address_field = models.GenericIPAddressField()
    url_field = models.URLField()
    float_field = models.FloatField()
    binary_field = models.BinaryField()
    boolean_field = models.BooleanField()
    decimal_field = models.DecimalField()
    time_field = models.TimeField()
    duration_field = models.DurationField()
    date_field = models.DateField()
    date_time_field = models.DateTimeField()
    uuid_field = models.UUIDField()
    tags_m2m = models.ManyToManyField(Tag)

    unidentifiable = NOT_FOUND

    #? models.IntegerField()
    integer_field

    def method(self):
        return 42

# -----------------
# Model attribute inference
# -----------------

#? DeferredAttribute()
BusinessModel.integer_field
#? DeferredAttribute()
BusinessModel.tags_m2m
#? DeferredAttribute()
BusinessModel.email_field

model_instance = BusinessModel()

#? int()
model_instance.integer_field
#? int()
model_instance.big_integer_field
#? int()
model_instance.positive_integer_field
#? int()
model_instance.small_integer_field
#? str()
model_instance.char_field
#? str()
model_instance.text_field
#? str()
model_instance.email_field
#? str()
model_instance.ip_address_field
#? str()
model_instance.url_field
#? float()
model_instance.float_field
#? bytes()
model_instance.binary_field
#? bool()
model_instance.boolean_field
#? decimal.Decimal()
model_instance.decimal_field
#? datetime.time()
model_instance.time_field
#? datetime.timedelta()
model_instance.duration_field
#? datetime.date()
model_instance.date_field
#? datetime.datetime()
model_instance.date_time_field
#? uuid.UUID()
model_instance.uuid_field

#! ['attached_o2o = models.OneToOneField(AttachedData)']
model_instance.attached_o2o
#! ['extra_data = models.TextField()']
model_instance.attached_o2o.extra_data
#? AttachedData()
model_instance.attached_o2o
#? str()
model_instance.attached_o2o.extra_data

#! ['category_fk = models.ForeignKey(Category)']
model_instance.category_fk
#! ['category_name = models.CharField()']
model_instance.category_fk.category_name
#? Category()
model_instance.category_fk
#? str()
model_instance.category_fk.category_name
#? Category()
model_instance.category_fk2
#? str()
model_instance.category_fk2.category_name
#?
model_instance.category_fk3
#?
model_instance.category_fk4
#?
model_instance.category_fk5

#? models.manager.RelatedManager()
model_instance.tags_m2m
#? Tag()
model_instance.tags_m2m.get()
#? ['add']
model_instance.tags_m2m.add

#?
model_instance.unidentifiable
#! ['unidentifiable = NOT_FOUND']
model_instance.unidentifiable

#? int()
model_instance.method()
#! ['def method']
model_instance.method

# -----------------
# Queries
# -----------------

#? ['objects']
model_instance.object
#?
model_instance.objects
#?
model_instance.objects.filter
#? models.query.QuerySet.filter
BusinessModel.objects.filter
#? BusinessModel() None
BusinessModel.objects.filter().first()
#? str()
BusinessModel.objects.get().char_field
#? int()
BusinessModel.objects.update(x='')
#? BusinessModel()
BusinessModel.objects.create()

# -----------------
# Custom object manager
# -----------------

#? TagManager()
Tag.objects
#? Tag() None
Tag.objects.filter().first()

#? TagManager()
Tag.custom_objects
#? Tag() None
Tag.custom_objects.filter().first()

# -----------------
# Inheritance
# -----------------

class Inherited(BusinessModel):
    text_field = models.IntegerField()
    new_field = models.FloatField()

inherited = Inherited()
#? int()
inherited.text_field
#? str()
inherited.char_field
#? float()
inherited.new_field

#?
Inherited.category_fk2.category_name
#? str()
inherited.category_fk2.category_name
#? str()
Inherited.objects.get().char_field
#? int()
Inherited.objects.get().text_field
#? float()
Inherited.objects.get().new_field

# -----------------
# Model methods
# -----------------

#? ['from_db']
Inherited.from_db
#? ['validate_unique']
Inherited.validate_uniqu
#? ['validate_unique']
Inherited().validate_unique

# -----------------
# Django Auth
# -----------------

#? str()
User().email
#? str()
User.objects.get().email

# -----------------
# values & values_list (dave is too lazy to implement it)
# -----------------

#?
BusinessModel.objects.values_list('char_field')[0]
#? dict()
BusinessModel.objects.values('char_field')[0]
#?
BusinessModel.objects.values('char_field')[0]['char_field']

# -----------------
# Completion
# -----------------

#? 19 ['text_field=']
Inherited(text_fiel)
#? 18 ['new_field=']
Inherited(new_fiel)
#? 19 ['char_field=']
Inherited(char_fiel)
#? 19 ['email_field=']
Inherited(email_fie)
#? 19 []
Inherited(unidentif)
#? 21 ['category_fk=', 'category_fk2=', 'category_fk3=', 'category_fk4=', 'category_fk5=']
Inherited(category_fk)
#? 21 ['attached_o2o=']
Inherited(attached_o2)
#? 18 ['tags_m2m=']
Inherited(tags_m2m)

#? 32 ['tags_m2m=']
Inherited.objects.create(tags_m2)
#? 32 ['tags_m2m=']
Inherited.objects.filter(tags_m2)
#? 35 ['char_field=']
Inherited.objects.exclude(char_fiel)
#? 34 ['char_field=']
Inherited.objects.update(char_fiel)
#? 32 ['email_field=']
Inherited.objects.get(email_fiel)
#? 44 ['category_fk2=']
Inherited.objects.get_or_create(category_fk2)
#? 44 ['uuid_field=']
Inherited.objects.update_or_create(uuid_fiel)
#? 48 ['char_field=']
Inherited.objects.exclude(pk=3).filter(char_fiel)



import threading
from typing import Any, Optional

SUBDEBUG: Any
SUBWARNING: Any

def sub_debug(msg, *args): ...
def debug(msg, *args): ...
def info(msg, *args): ...
def sub_warning(msg, *args): ...
def get_logger(): ...
def log_to_stderr(level: Optional[Any] = ...): ...
def get_temp_dir(): ...
def register_after_fork(obj, func): ...

class Finalize:
    def __init__(self, obj, callback, args=..., kwargs: Optional[Any] = ..., exitpriority: Optional[Any] = ...): ...
    def __call__(self, wr: Optional[Any] = ...): ...
    def cancel(self): ...
    def still_active(self): ...

def is_exiting(): ...

class ForkAwareThreadLock:
    def __init__(self): ...

class ForkAwareLocal(threading.local):
    def __init__(self): ...
    def __reduce__(self): ...



from os.path import join

import pytest
import py

from ..helpers import get_example_dir, example_dir
from jedi import Project


SYS_PATH = [get_example_dir('namespace_package', 'ns1'),
            get_example_dir('namespace_package', 'ns2')]


def script_with_path(Script, *args, **kwargs):
    return Script(project=Project('.', sys_path=SYS_PATH), *args, **kwargs)


def test_goto_definition(Script):
    assert script_with_path(Script, 'from pkg import ns1_file').infer()
    assert script_with_path(Script, 'from pkg import ns2_file').infer()
    assert not script_with_path(Script, 'from pkg import ns3_file').infer()


@pytest.mark.parametrize(
    ('source', 'solution'), [
        ('from pkg.ns2_folder.nested import foo', 'nested!'),
        ('from pkg.ns2_folder import foo', 'ns2_folder!'),
        ('from pkg.ns2_file import foo', 'ns2_file!'),
        ('from pkg.ns1_folder import foo', 'ns1_folder!'),
        ('from pkg.ns1_file import foo', 'ns1_file!'),
        ('from pkg import foo', 'ns1!'),
    ]
)
def test_goto_assignment(Script, source, solution):
    ass = script_with_path(Script, source).goto()
    assert len(ass) == 1
    assert ass[0].description == "foo = '%s'" % solution


def test_simple_completions(Script):
    # completion
    completions = script_with_path(Script, 'from pkg import ').complete()
    names = [c.name for c in completions]
    compare = ['foo', 'ns1_file', 'ns1_folder', 'ns2_folder', 'ns2_file',
               'pkg_resources', 'pkgutil', '__name__', '__path__',
               '__package__', '__file__', '__doc__']
    # must at least contain these items, other items are not important
    assert set(compare) == set(names)


@pytest.mark.parametrize(
    ('source', 'solution'), [
        ('from pkg import ns2_folder as x', 'ns2_folder!'),
        ('from pkg import ns2_file as x', 'ns2_file!'),
        ('from pkg.ns2_folder import nested as x', 'nested!'),
        ('from pkg import ns1_folder as x', 'ns1_folder!'),
        ('from pkg import ns1_file as x', 'ns1_file!'),
        ('import pkg as x', 'ns1!'),
    ]
)
def test_completions(Script, source, solution):
    for c in script_with_path(Script, source + '; x.').complete():
        if c.name == 'foo':
            completion = c
    solution = "foo = '%s'" % solution
    assert completion.description == solution


def test_nested_namespace_package(Script):
    code = 'from nested_namespaces.namespace.pkg import CONST'

    sys_path = [example_dir]
    project = Project('.', sys_path=sys_path)
    result = Script(code, project=project).infer(line=1, column=45)

    assert len(result) == 1


def test_relative_import(Script, environment, tmpdir):
    """
    Attempt a relative import in a very simple namespace package.
    """
    directory = get_example_dir('namespace_package_relative_import')
    # Need to copy the content in a directory where there's no __init__.py.
    py.path.local(directory).copy(tmpdir)
    file_path = join(tmpdir.strpath, "rel1.py")
    script = Script(path=file_path)
    d, = script.infer(line=1)
    assert d.name == 'int'
    d, = script.goto(line=1)
    assert d.name == 'name'
    assert d.module_name == 'rel2'
